@@ -14,7 +14,16 @@ from sonolus.backend.utils import get_function, scan_writes
 from sonolus.script.debug import assert_true
 from sonolus.script.internal.builtin_impls import BUILTIN_IMPLS, _bool, _float, _int, _len, _super
 from sonolus.script.internal.constant import ConstantValue
-from sonolus.script.internal.context import Context, EmptyBinding, Scope, ValueBinding, ctx, set_ctx, using_ctx
+from sonolus.script.internal.context import (
+    ConflictBinding,
+    Context,
+    EmptyBinding,
+    Scope,
+    ValueBinding,
+    ctx,
+    set_ctx,
+    using_ctx,
+)
 from sonolus.script.internal.descriptor import SonolusDescriptor
 from sonolus.script.internal.error import CompilationError
 from sonolus.script.internal.impl import meta_fn, validate_value
@@ -22,7 +31,7 @@ from sonolus.script.internal.transient import TransientValue
 from sonolus.script.internal.tuple_impl import TupleImpl
 from sonolus.script.internal.value import Value
 from sonolus.script.iterator import SonolusIterator
-from sonolus.script.maybe import Maybe
+from sonolus.script.maybe import Maybe, Nothing
 from sonolus.script.num import Num, _is_num
 from sonolus.script.record import Record
 
@@ -38,9 +47,14 @@ def compile_and_call[**P, R](fn: Callable[P, R], /, *args: P.args, **kwargs: P.k
 def compile_and_call_at_definition[**P, R](fn: Callable[P, R], /, *args: P.args, **kwargs: P.kwargs) -> R:
     if not ctx():
         return fn(*args, **kwargs)
-    if ctx().no_eval:
-        return compile_and_call(fn, *args, **kwargs)
     source_file, node = get_function(fn)
+    if ctx().no_eval:
+        debug_stack = ctx().callback_state.debug_stack
+        try:
+            debug_stack.append(f'File "{source_file}", line {node.lineno}, in <callback>')
+            return compile_and_call(fn, *args, **kwargs)
+        finally:
+            debug_stack.pop()
     location_args = {
         "lineno": node.lineno,
         "col_offset": node.col_offset,
@@ -55,10 +69,16 @@ def compile_and_call_at_definition[**P, R](fn: Callable[P, R], /, *args: P.args,
             **location_args,
         )
     )
-    return eval(
-        compile(expr, filename=source_file, mode="eval"),
-        {"fn": lambda: compile_and_call(fn, *args, **kwargs), "_filter_traceback_": True, "_traceback_root_": True},
-    )
+
+    debug_stack = ctx().callback_state.debug_stack
+    try:
+        debug_stack.append(f'File "{source_file}", line {node.lineno}, in <callback>')
+        return eval(
+            compile(expr, filename=source_file, mode="eval"),
+            {"fn": lambda: compile_and_call(fn, *args, **kwargs), "_filter_traceback_": True, "_traceback_root_": True},
+        )
+    finally:
+        debug_stack.pop()
 
 
 def generate_fn_impl(fn: Callable):
@@ -102,7 +122,9 @@ def eval_fn(fn: Callable, /, *args, **kwargs):
         **fn.__globals__,
         **nonlocal_vars,
     }
-    return Visitor(source_file, bound_args, global_vars).run(node)
+    return Visitor(
+        source_file, bound_args, global_vars, parent=None, function_name=getattr(fn, "__name__", "<unnamed>")
+    ).run(node)
 
 
 unary_ops = {
@@ -223,14 +245,16 @@ class Visitor(ast.NodeVisitor):
     resume_ctxs: list[Context]  # Contexts after yield statements
     active_ctx: Context | None  # The active context for use in nested functions
     parent: Visitor | None  # The parent visitor for use in nested functions
-    used_parent_binding_values: dict[str, Value]  # Values of parent bindings used in this function
+    used_parent_binding_values: dict[str, Value]  # Values of parent bindings used in this
+    function_name: str
 
     def __init__(
         self,
         source_file: str,
         bound_args: inspect.BoundArguments,
         global_vars: dict[str, Any],
-        parent: Visitor | None = None,
+        parent: Visitor | None,
+        function_name: str,
     ):
         self.source_file = source_file
         self.globals = global_vars
@@ -244,6 +268,7 @@ class Visitor(ast.NodeVisitor):
         self.active_ctx = None
         self.parent = parent
         self.used_parent_binding_values = {}
+        self.function_name = function_name
 
     def run(self, node):
         before_ctx = ctx()
@@ -312,13 +337,16 @@ class Visitor(ast.NodeVisitor):
                 yield_merge_ctx = Context.meet(yield_between_ctxs)
             else:
                 yield_merge_ctx = before_ctx.new_empty_disconnected()
-                # Making it default to a number for convenience when used with stuff like min, etc.
-                yield_merge_ctx.scope.set_value("$yield", validate_value(0))
             yield_binding = yield_merge_ctx.scope.get_binding("$yield")
-            if not isinstance(yield_binding, ValueBinding):
-                raise ValueError("Function has conflicting yield values")
-            with using_ctx(yield_merge_ctx):
-                is_present_var._set_(1)
+            match yield_binding:
+                case ValueBinding():
+                    with using_ctx(yield_merge_ctx):
+                        is_present_var._set_(1)
+                    yield_value = Maybe(present=is_present_var, value=yield_binding.value)
+                case EmptyBinding():
+                    yield_value = Nothing
+                case ConflictBinding():
+                    raise ValueError("Function has conflicting yield values")
             next_result_ctx = Context.meet([yield_merge_ctx, return_ctx])
             set_ctx(before_ctx)
             return_test = Num._alloc_()
@@ -327,7 +355,7 @@ class Visitor(ast.NodeVisitor):
                 return_test,
                 entry,
                 next_result_ctx,
-                Maybe(present=is_present_var, value=yield_binding.value),
+                yield_value,
                 self.used_parent_binding_values,
                 self,
             )
@@ -421,7 +449,8 @@ class Visitor(ast.NodeVisitor):
                 self.source_file,
                 bound,
                 self.globals,
-                self,
+                parent=self,
+                function_name=name,
             ).run(node)
 
         fn._meta_fn_ = True
@@ -739,11 +768,17 @@ class Visitor(ast.NodeVisitor):
                         test = validate_value(subject._is_py_() and subject._as_py_() is None)
                     case _:
                         raise NotImplementedError("Unsupported match singleton")
-                ctx_init = ctx()
-                ctx_init.test = test.ir()
-                true_ctx = ctx_init.branch(None)
-                false_ctx = ctx_init.branch(0)
-                return true_ctx, false_ctx
+                if test._is_py_():
+                    if test._as_py_():
+                        return ctx(), ctx().into_dead()
+                    else:
+                        return ctx().into_dead(), ctx()
+                else:
+                    ctx_init = ctx()
+                    ctx_init.test = test.ir()
+                    true_ctx = ctx_init.branch(None)
+                    false_ctx = ctx_init.branch(0)
+                    return true_ctx, false_ctx
             case ast.MatchSequence(patterns=patterns):
                 target_len = len(patterns)
                 if not (isinstance(subject, Sequence | TupleImpl)):
@@ -933,7 +968,8 @@ class Visitor(ast.NodeVisitor):
                 self.source_file,
                 bound,
                 self.globals,
-                self,
+                parent=self,
+                function_name="<lambda>",
             ).run(node)
 
         fn._meta_fn_ = True
@@ -984,7 +1020,9 @@ class Visitor(ast.NodeVisitor):
 
     def visit_GeneratorExp(self, node):
         self.active_ctx = ctx()
-        return Visitor(self.source_file, inspect.Signature([]).bind(), self.globals, self).run(node)
+        return Visitor(
+            self.source_file, inspect.Signature([]).bind(), self.globals, parent=self, function_name="<genexp>"
+        ).run(node)
 
     def visit_Await(self, node):
         raise NotImplementedError("Await expressions are not supported")
@@ -1308,15 +1346,20 @@ class Visitor(ast.NodeVisitor):
     ) -> R | Value:
         """Handles a call to the given callable."""
         self.active_ctx = ctx()
-        if (
-            isinstance(fn, Value)
-            and fn._is_py_()
-            and isinstance(fn._as_py_(), type)
-            and issubclass(fn._as_py_(), Value)
-        ):
-            return validate_value(self.execute_at_node(node, fn._as_py_(), *args, **kwargs))
-        else:
-            return self.execute_at_node(node, lambda: validate_value(compile_and_call(fn, *args, **kwargs)))
+        debug_stack = self.active_ctx.callback_state.debug_stack
+        try:
+            debug_stack.append(f'File "{self.source_file}", line {node.lineno}, in {self.function_name}')
+            if (
+                isinstance(fn, Value)
+                and fn._is_py_()
+                and isinstance(fn._as_py_(), type)
+                and issubclass(fn._as_py_(), Value)
+            ):
+                return validate_value(self.execute_at_node(node, fn._as_py_(), *args, **kwargs))
+            else:
+                return self.execute_at_node(node, lambda: validate_value(compile_and_call(fn, *args, **kwargs)))
+        finally:
+            debug_stack.pop()
 
     def handle_getitem(self, node: ast.stmt | ast.expr, target: Value, key: Value) -> Value:
         with self.reporting_errors_at_node(node):
