@@ -11,7 +11,7 @@ from sonolus.script.debug import assert_unreachable
 from sonolus.script.internal.context import ctx
 from sonolus.script.internal.error import InternalError
 from sonolus.script.internal.generic import GenericValue
-from sonolus.script.internal.impl import meta_fn, validate_value
+from sonolus.script.internal.impl import meta_fn, perf_meta_fn, validate_value
 from sonolus.script.internal.value import BackingSource, DataValue, Value
 from sonolus.script.num import Num
 
@@ -188,12 +188,28 @@ class Array[T, Size](GenericValue, ArrayLike[T], metaclass=ArrayMeta):
         else:
             return cls._with_value([cls.element_type()._zero_() for _ in range(cls.size())])
 
+    @perf_meta_fn
     def __len__(self):
         return self.size()
 
     @meta_fn
     def __getitem__(self, index: int) -> T:
-        index: Num = Num._accept_(get_positive_index(index, self.size()))
+        return self.get_unchecked(get_positive_index(index, self.size()))
+
+    @meta_fn
+    def get_unchecked(self, index: Num) -> T:
+        """Get the element at the given index possibly without bounds checking.
+
+        The compiler may still determine that the index is out of bounds and throw an error, but it may skip these
+        checks at runtime.
+
+        Args:
+            index: The index to get.
+
+        Returns:
+            The element at the given index.
+        """
+        index = Num._accept_(index)
         if index._is_py_() and 0 <= index._as_py_() < self.size():
             const_index = index._as_py_()
             if isinstance(const_index, float) and not const_index.is_integer():
@@ -212,7 +228,8 @@ class Array[T, Size](GenericValue, ArrayLike[T], metaclass=ArrayMeta):
                 )
             elif callable(self._value):
                 return self.element_type()._from_backing_source_(
-                    lambda offset: self._value((Num(offset) + Num(const_index * self.element_type()._size_())).ir())  # type: ignore
+                    lambda offset: self._value((Num(offset) + Num(const_index * self.element_type()._size_())).ir())
+                    # type: ignore
                 )
             else:
                 raise InternalError("Unexpected array value")
@@ -237,7 +254,20 @@ class Array[T, Size](GenericValue, ArrayLike[T], metaclass=ArrayMeta):
 
     @meta_fn
     def __setitem__(self, index: int, value: T):
-        index: Num = Num._accept_(get_positive_index(index, self.size()))
+        self.set_unchecked(get_positive_index(index, self.size()), value)
+
+    @meta_fn
+    def set_unchecked(self, index: Num, value: T):
+        """Set the element at the given index possibly without bounds checking.
+
+        The compiler may still determine that the index is out of bounds and throw an error, but it may skip these
+        checks at runtime.
+
+        Args:
+            index: The index to set.
+            value: The value to set.
+        """
+        index = Num._accept_(index)
         value = self.element_type()._accept_(value)
         if ctx():
             if isinstance(self._value, list):
@@ -287,7 +317,7 @@ class Array[T, Size](GenericValue, ArrayLike[T], metaclass=ArrayMeta):
             return False
         i = 0
         while i < self.size():
-            if self[i] != other[i]:
+            if self.get_unchecked(i) != other.get_unchecked(i):
                 return False
             i += 1
         return True
