@@ -28,7 +28,7 @@ class Project:
         engine: Engine,
         levels: Iterable[Level] | Callable[[], Iterable[Level]] | None = None,
         resources: PathLike | None = None,
-        converters: dict[str | None, Callable[[ExternalLevelData], LevelData]] | None = None,
+        converters: dict[str | None, Callable[[ExternalLevelData], LevelData | None]] | None = None,
     ):
         self.engine = engine
         match levels:
@@ -63,10 +63,20 @@ class Project:
             port: The port of the development server.
             config: The build configuration.
         """
-        from sonolus.build.cli import build_collection, run_server
+        from sonolus.build.cli import run_server
 
-        build_collection(self, Path(build_dir), config)
-        run_server(Path(build_dir) / "site", port=port)
+        if config is None:
+            config = BuildConfig()
+
+        run_server(
+            Path(build_dir) / "site",
+            port=port,
+            project_module_name=None,
+            core_module_names=None,
+            build_dir=Path(build_dir),
+            config=config,
+            project=self,
+        )
 
     def build(self, build_dir: PathLike, config: BuildConfig | None = None):
         """Build the project.
