@@ -1,10 +1,15 @@
+from collections.abc import Iterable
 from dataclasses import dataclass
 from enum import StrEnum
 from typing import Annotated, Any, NewType, dataclass_transform, get_origin
 
 from sonolus.backend.ops import Op
+from sonolus.script.array_like import ArrayLike, check_positive_index
+from sonolus.script.debug import static_error
+from sonolus.script.internal.impl import perf_meta_fn
 from sonolus.script.internal.introspection import get_field_specifiers
 from sonolus.script.internal.native import native_function
+from sonolus.script.num import Num
 from sonolus.script.quad import QuadLike, flatten_quad
 from sonolus.script.record import Record
 from sonolus.script.vec import Vec2
@@ -26,6 +31,7 @@ class Sprite(Record):
         """Check if the sprite is available."""
         return _has_skin_sprite(self.id)
 
+    @perf_meta_fn
     def draw(self, quad: QuadLike, z: float = 0.0, a: float = 1.0):
         """Draw the sprite.
 
@@ -36,6 +42,7 @@ class Sprite(Record):
         """
         _draw(self.id, *flatten_quad(quad), z, a)
 
+    @perf_meta_fn
     def draw_curved_b(self, quad: QuadLike, cp: Vec2, n: float, z: float = 0.0, a: float = 1.0):
         """Draw the sprite with a curved bottom with a quadratic Bézier curve.
 
@@ -48,6 +55,7 @@ class Sprite(Record):
         """
         _draw_curved_b(self.id, *flatten_quad(quad), z, a, n, *cp.tuple)
 
+    @perf_meta_fn
     def draw_curved_t(self, quad: QuadLike, cp: Vec2, n: float, z: float = 0.0, a: float = 1.0):
         """Draw the sprite with a curved top with a quadratic Bézier curve.
 
@@ -60,6 +68,7 @@ class Sprite(Record):
         """
         _draw_curved_t(self.id, *flatten_quad(quad), z, a, n, *cp.tuple)
 
+    @perf_meta_fn
     def draw_curved_l(self, quad: QuadLike, cp: Vec2, n: float, z: float = 0.0, a: float = 1.0):
         """Draw the sprite with a curved left side with a quadratic Bézier curve.
 
@@ -72,6 +81,7 @@ class Sprite(Record):
         """
         _draw_curved_l(self.id, *flatten_quad(quad), z, a, n, *cp.tuple)
 
+    @perf_meta_fn
     def draw_curved_r(self, quad: QuadLike, cp: Vec2, n: float, z: float = 0.0, a: float = 1.0):
         """Draw the sprite with a curved right side with a quadratic Bézier curve.
 
@@ -84,6 +94,7 @@ class Sprite(Record):
         """
         _draw_curved_r(self.id, *flatten_quad(quad), z, a, n, *cp.tuple)
 
+    @perf_meta_fn
     def draw_curved_bt(self, quad: QuadLike, cp1: Vec2, cp2: Vec2, n: float, z: float = 0.0, a: float = 1.0):
         """Draw the sprite with a curved bottom and top with a cubic Bézier curve.
 
@@ -97,6 +108,7 @@ class Sprite(Record):
         """
         _draw_curved_bt(self.id, *flatten_quad(quad), z, a, n, *cp1.tuple, *cp2.tuple)
 
+    @perf_meta_fn
     def draw_curved_lr(self, quad: QuadLike, cp1: Vec2, cp2: Vec2, n: float, z: float = 0.0, a: float = 1.0):
         """Draw the sprite with a curved left and right side with a cubic Bézier curve.
 
@@ -111,6 +123,32 @@ class Sprite(Record):
         _draw_curved_lr(self.id, *flatten_quad(quad), z, a, n, *cp1.tuple, *cp2.tuple)
 
 
+class SpriteGroup(Record, ArrayLike[Sprite]):
+    """A group of sprites.
+
+    Usage:
+        ```python
+        SpriteGroup(start_id: int, size: int)
+        ```
+    """
+
+    start_id: int
+    size: int
+
+    def __len__(self) -> int:
+        return self.size
+
+    def __getitem__(self, index: int) -> Sprite:
+        check_positive_index(index, self.size)
+        return Sprite(self.start_id + index)
+
+    def get_unchecked(self, index: Num) -> Sprite:
+        return Sprite(self.start_id + index)
+
+    def __setitem__(self, index: int, value: Sprite) -> None:
+        static_error("SpriteGroup is read-only")
+
+
 @native_function(Op.HasSkinSprite)
 def _has_skin_sprite(sprite_id: int) -> bool:
     raise NotImplementedError
@@ -262,11 +300,21 @@ class SkinSprite:
     name: str
 
 
+@dataclass
+class SkinSpriteGroup:
+    names: list[str]
+
+
 def sprite(name: str) -> Any:
     """Define a sprite with the given name."""
     return SkinSprite(name)
 
 
+def sprite_group(names: Iterable[str]) -> Any:
+    """Define a sprite group with the given names."""
+    return SkinSpriteGroup(list(names))
+
+
 type Skin = NewType("Skin", Any)  # type: ignore
 
 
@@ -294,25 +342,44 @@ def skin[T](cls: type[T]) -> T | Skin:
             render_mode: RenderMode = RenderMode.LIGHTWEIGHT
 
             note: StandardSprite.NOTE_HEAD_RED
-            other: Sprite = skin_sprite("other")
+            other: Sprite = sprite("other")
+            group_1: SpriteGroup = sprite_group(["one", "two", "three"])
+            group_2: SpriteGroup = sprite_group(f"name_{i}" for i in range(10))
         ```
     """
     if len(cls.__bases__) != 1:
         raise ValueError("Skin class must not inherit from any class (except object)")
     instance = cls()
     names = []
-    for i, (name, annotation) in enumerate(get_field_specifiers(cls, skip={"render_mode"}).items()):
+    i = 0
+    for name, annotation in get_field_specifiers(cls, skip={"render_mode"}).items():
         if get_origin(annotation) is not Annotated:
-            raise TypeError(f"Invalid annotation for skin: {annotation}")
+            raise TypeError(f"Invalid annotation for skin: {annotation} on field {name}")
         annotation_type = annotation.__args__[0]
         annotation_values = annotation.__metadata__
-        if annotation_type is not Sprite:
-            raise TypeError(f"Invalid annotation for skin: {annotation}, expected annotation of type Sprite")
-        if len(annotation_values) != 1 or not isinstance(annotation_values[0], SkinSprite):
-            raise TypeError(f"Invalid annotation for skin: {annotation}, expected a single string annotation value")
-        sprite_name = annotation_values[0].name
-        names.append(sprite_name)
-        setattr(instance, name, Sprite(i))
+        if len(annotation_values) != 1:
+            raise TypeError(f"Invalid annotation for skin: {annotation} on field {name}, too many annotation values")
+        sprite_info = annotation_values[0]
+        match sprite_info:
+            case SkinSprite(name=sprite_name):
+                if annotation_type is not Sprite:
+                    raise TypeError(f"Invalid annotation for skin: {annotation} on field {name}, expected Sprite")
+                names.append(sprite_name)
+                setattr(instance, name, Sprite(i))
+                i += 1
+            case SkinSpriteGroup(names=sprite_names):
+                if annotation_type is not SpriteGroup:
+                    raise TypeError(f"Invalid annotation for skin: {annotation} on field {name}, expected SpriteGroup")
+                start_id = i
+                count = len(sprite_names)
+                names.extend(sprite_names)
+                setattr(instance, name, SpriteGroup(start_id, count))
+                i += count
+            case _:
+                raise TypeError(
+                    f"Invalid annotation for skin: {annotation} on field {name}, unknown sprite info, "
+                    f"expected a skin() or sprite_group() specifier"
+                )
     instance._sprites_ = names
     instance.render_mode = RenderMode(getattr(instance, "render_mode", RenderMode.DEFAULT))
     instance._is_comptime_value_ = True
